from typing import Generic, Iterable, TypeAlias, TypeVar

from rainbowadn.core import *
from rainbowadn.flow.verification.core import *
from rainbowadn.nullability import *
from ._flowstandard import *

__all__ = ('FlowBlock', 'FlowBlockFactory', 'FlowBlockVerification',)

LinkT = TypeVar('LinkT')

FBL: TypeAlias = 'FlowBlock[LinkT]'
Index: TypeAlias = FlowStandard[FBL]


class FlowBlock(Generic[LinkT], RecursiveMentionable):
    def points(self) -> Iterable[HashPoint]:
        return [*self.previous.points(), *self.index.points(), self.link]

    def __bytes__(self):
        return bytes(self.previous) + bytes(self.index) + bytes(self.link)

    def __factory__(self) -> RainbowFactory[FBL]:
        return FlowBlockFactory(self.link.factory)

    def __init__(
            self,
            previous: NullableReference[FBL],
            index: Index,
            link: HashPoint[LinkT],
    ):
        assert isinstance(previous, NullableReference)
        assert isinstance(index, FlowStandard)
        self.previous = previous
        self.index = index
        self.link = link

    async def outer(self) -> Index:
        return FlowStandard(
            (
                await self.index.protocolized.tree.add(
                    HashPoint.of(self)
                )
            ).protocolized()
        )

    @classmethod
    async def outer_of(cls, factory: RainbowFactory[LinkT], reference: NullableReference[FBL]) -> Index:
        if reference.null():
            return FlowStandardFactory.empty(FlowBlockFactory(factory))
        else:
            return await (await reference.resolve()).outer()

    @classmethod
    async def link_of(cls, reference: NullableReference[FBL]) -> Nullable[LinkT]:
        if reference.null():
            return Null()
        else:
            return NotNull(await (await reference.resolve()).link.resolve())

    async def add(self, link: HashPoint[LinkT]) -> FBL:
        return FlowBlock(
            NullableReference.off(self),
            await self.outer(),
            link
        )

    @classmethod
    async def add_to(cls, reference: NullableReference[FBL], link: HashPoint[LinkT]) -> FBL:
        return FlowBlock(
            reference,
            await cls.outer_of(link.factory, reference),
            link
        )

    async def str(self, tab: int) -> str:
        assert isinstance(tab, int)
        previous_str: str
        link_str: str
        previous_str, link_str = await gather(
            self.previous.str(tab),
            hash_point_format(self.link, tab)
        )
        assert isinstance(previous_str, str)
        assert isinstance(link_str, str)
        return f'{previous_str}' \
               f'{tabulate(tab)}(index)' \
               f'{tabulate(tab)}{link_str}'


class FlowBlockFactory(RainbowFactory[FBL], Generic[LinkT]):
    def __init__(self, factory: RainbowFactory[LinkT]):
        assert isinstance(factory, RainbowFactory)
        self.factory = factory

    def from_bytes(self, source: bytes, resolver: HashResolver) -> FBL:
        assert isinstance(source, bytes)
        assert isinstance(resolver, HashResolver)
        return FlowBlock(
            NullableReferenceFactory(self).from_bytes(source[:HashPoint.HASH_LENGTH], resolver),
            FlowStandardFactory.of(self).from_bytes(source[HashPoint.HASH_LENGTH:2 * HashPoint.HASH_LENGTH], resolver),
            ResolverOrigin(self.factory, source[2 * HashPoint.HASH_LENGTH:], resolver).hash_point(),
        )

    def loose(self) -> RainbowFactory[FBL]:
        return self


class FlowBlockIndexedVerification(
    Verification[FBL],
    Generic[LinkT]
):
    def __init__(
            self,
            index: Index,
            verification: Verification[tuple[Nullable[LinkT], LinkT]],
    ):
        assert isinstance(index, FlowStandard)
        assert isinstance(verification, Verification)
        self.index = index
        self.verification = verification

    async def verify(self, element: FBL) -> bool:
        link = await element.link.resolve()
        if element.previous.null():
            assert_trues(
                await gather(
                    self.verification.verify((Null(), link)),
                    element.index.verify_empty(),
                )
            )
        else:
            previous: FBL = await element.previous.resolve()

            async def verify_link() -> bool:
                assert_true(await self.verification.verify((NotNull(await previous.link.resolve()), link)))
                return True

            async def verify_contains() -> bool:
                assert_true(await self.index.contains(previous))
                return True

            async def verify_index() -> bool:
                assert_eq(element.index, await previous.outer())
                return True

            assert_trues(
                await gather(
                    verify_link(),
                    verify_contains(),
                    verify_index(),
                )
            )
        return True

    def loose(self) -> Verification[FBL]:
        return self


class FlowBlockVerification(
    Verification[Index],
    Generic[LinkT]
):
    def __init__(
            self,
            verification: Verification[tuple[Nullable[LinkT], LinkT]],
    ):
        assert isinstance(verification, Verification)
        self.verification = verification

    async def verify(self, element: Index) -> bool:
        assert_true(
            await ReduceVerification(
                FlowBlockIndexedVerification(element, self.verification)
            ).verify(
                await element.reducer()
            )
        )
        return True

    def loose(self) -> Verification[Index]:
        return self
